# coding: utf-8

'''
Python3 code to calculate strain and velocity from fault source parameters.

If using this code, please cite the accompanying paper:
Reitman, N.R. and P.H. Molnar (2020) Strain and velocity across the Great Basin
derived from 15-ka fault slip rates: Implications for continuous deformation
and seismic hazard in the Walker Lane.


Fault Data:
The input fault source parameter data are in a GeoPandas database called “hazfaults”.
Required inputs are fault strike, dip, rake, length, and slip rate.
Data used in the above paper are available as a supplement to the paper
or in this Zenodo archive under DOI: 10.5281/zenodo.3895334


Shapefile Dependencies:
states: https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html
Great Basin: http://www.nbmg.unr.edu/Geothermal/Data.html


Software Dependencies:
Python3
NumPy
Pickle
Pandas
GeoPandas
Shapely
EarthPy: https://earthpy.readthedocs.io/en/latest/index.html

please contact nadine.reitman@colorado.edu with questions or comments.

'''


###############
#### SETUP ####
###############

import numpy as np
from shapely.geometry import Polygon, Point, LineString, MultiLineString, mapping
import pickle
import geopandas as gpd
import pandas as pd
from earthpy import clip


############################
#### LOAD FAULT DATA #######
############################

# load data with strike, dip, rake
hazfaults = pickle.load(open('hazfaults_clean.p','rb'))

# copy rate column to vtotal for easier coding
hazfaults['vtotal'] = hazfaults.rate.copy()

# rename wassuk normal b/c it's misspelled
hazfaults.loc[186,'name'] = 'Wassuk Range (normal)'

# make vertical slip rates the same as other slip rates
for index,row in hazfaults.iterrows():
    if row['rateType'] == 'VERTICAL_SLIP_RATE':
        hazfaults.loc[index,'vtotal'] = row['rate']/np.sin(np.radians(row['dip']))


############################
#### LOAD SHAPEFILES    ####
############################

### Load States Shapefile ###
# read in states with geopandas
states = gpd.read_file('/Users/nadine/Documents/Data/USA/vector/boundaries-conus/usa-states-census-2014.shp')
print('CRS of states shapefile: %s'%states.crs)

# make utm version
states_utm = states.to_crs({'init': 'epsg:32611'})
print('CRS of reprojected states shapefile: %s' %states_utm.crs)


### Load Great Basin Shapefile ###
# read in GB polygon with geopandas
greatbasin = gpd.read_file('/Users/nadine/Documents/Data/DeathValley/vector/GreatBasinOutline/GreatBasinOutline.shp')
print('CRS of Great Basin shapefile: %s' %greatbasin.crs)

# reproject to same crs as states
greatbasin_utm = greatbasin.to_crs({'init': 'epsg:32611'})
print('CRS of reprojected Great Basin shapefile: %s' %greatbasin_utm.crs)


## calculate strain and velocity

######################
### SET PARAMETERS ###
######################
strike = 317. # orientation of coordinate system
step = 25 # swath/bin width in km

###########################################
### CALCULATE DISTANCE TO SIERRA NEVADA ###
###########################################

### orientation set above, get slope
slope = np.round(np.tan(np.radians(90-strike)),3)

### get SN line equation ###
## need one point (y0,x0) that falls on the fault in order to calculate b, the y-intercept of the fault
y0,x0 = 37.992104, -119.143491 # for sierra nevada, this is centrally located near Mono Lake

# get this point in utm for plotting on map
easting = 311776.30
northing = 4207106.76

## y = mx + b --> b = y - mx
b = y0 - (slope*x0)

# do dist calc w/ lat/long
# distance, d = (b + (m*x0) - y0) / sqrt(1 + m**2)
x = hazfaults['x_lon']
y = hazfaults['y_lat']

dist = (b + (slope*x) - y) / np.sqrt(1 + (slope**2)) # this is in lat/lon degrees b/c we did equation of line in lat/lon degrees
dist = dist * np.cos(np.radians(y)) * -111.321 # convert from lat/lon degrees to km. length of 1 degree longitude at equator is 111.321 km.

# add distance in km to fault database
hazfaults['SNdist_calc'] = dist


################################
###  CALCULATE BIN LENGTHS  ####
################################
theta = np.radians(strike)
slope = np.tan(np.radians(-(360.-strike)))

### CLIP STATES SHAPEFILE ####
# clip states to greatbasin outline
states_clip = states_utm.copy()
states_clip = clip.clip_shp(states_clip,greatbasin_utm) # using the earthpy clip function

# dissolve states shapefile into 1 piece
states_clip = states_clip.dissolve(by='LSAD') # LSAD is just some column where they all have the same value

# clip north and south ends off
desired_box = Polygon([(0, 4670000), (1200000, 4670000), (1200000, 3930000), (0, 3930000)])
states_clip['geometry'] = states_clip.intersection(desired_box)
states_clip.crs = 'epsg:32611'

### set bid width ###
#step = 25 # bin width in km --> set above
bins = np.arange(-125,800,step) # make bins
width = step * 1000 # convert bin width from km to meters

# set min/max for long lines that will intercept w/ study area
below = 3850000 # any value south of study area
above = 4800000 # any value north of study area

############################
#### INITIALIZE ARRAYS  ####
############################

# initialize arrays to hold bin intercepts
areas = np.zeros(len(bins)-1)
lengths = np.zeros(len(bins)-1)
bottoms = np.zeros(len(bins)-1)
tops = np.zeros(len(bins)-1)

# initialize arrays to hold strains for each bin
summed_Enn = np.zeros(len(bins)-1)
summed_Eee = np.zeros(len(bins)-1)
summed_Een = np.zeros(len(bins)-1)

# initialize arrays to hold rotated strains for each bin
summed_Enn_rot = np.zeros(len(bins)-1)
summed_Eee_rot = np.zeros(len(bins)-1)
summed_Een_rot = np.zeros(len(bins)-1)

# intialize columns in hazfaults to hold strains
hazfaults['Eee_rot'] = 0
hazfaults['Enn_rot'] = 0
hazfaults['Een_rot'] = 0


################################
### CALCULATE STRAIN IN BINS ###
################################

# loop through each bin
for i in range(len(bins)-1):

    # select faults within this SNdist bin range
    selected = hazfaults[(hazfaults.SNdist_calc>bins[i]) & (hazfaults.SNdist_calc<=bins[i+1])].copy()

    # if there are faults in the bin, calculate area of the bin (area = 0 if no faults in bin)
    if len(selected)>0:

        # make a line for each bin and see where it intercepts w/ study area
        # first (middle) x coordinate is mean of all x of all faults in this bin
        center_x1 = np.nanmean(selected.x_utm)

        # second (bottom) x coordinate is determined by slope from strike
        center_x2 = -slope*(np.nanmean(selected.y_utm)-below) + center_x1

        # third (top) x coordinate is determined by slope from strike
        center_x3 = -slope*(below-above) + center_x2

        # make a line that extends above and below the study area
        A = Point(center_x3,above)
        B = Point(center_x2,below)
        AB = LineString([A,B])

        # find the intersect of the line with the study area
        intersect = LineString([A,B]).intersection(states_clip.geometry[0])

        # get the top, bottom, left, right coordinates of the intersection points
        # if only 2 intersections, the intersect is a LineString
        if type(intersect) == LineString:
            top = intersect.coords[0][1]
            bottom = intersect.coords[1][1]
            left = intersect.coords[0][0]
            right = intersect.coords[1][0]

        # if more than 2 intersects, the intersect is a MultiLineString
        elif type(intersect)==MultiLineString:
            top = mapping(intersect)['coordinates'][1][0][1]
            bottom = mapping(intersect)['coordinates'][1][1][1]
            left = mapping(intersect)['coordinates'][1][0][0]
            right = mapping(intersect)['coordinates'][1][1][0]

        # get bin length based on strike
        bin_length = (top - bottom) / np.cos(np.radians(360-strike))

        # get bin area
        area = width * bin_length

        # save length, area, top, bottom for later
        lengths[i] = bin_length # save length for lather
        areas[i] = area # save area for later
        tops[i] = top
        bottoms[i] = bottom

        ## calculate Mnn, Mee, Men for each fault
        Mnn = -(np.sin(np.radians(selected.dip))*np.cos(np.radians(selected.rake))*np.sin(np.radians(2*selected.strike)) + (np.sin(np.radians(2*selected.dip))*np.sin(np.radians(selected.rake))*(np.sin(np.radians(selected.strike)))**2))
        Mee = (np.sin(np.radians(selected.dip))*np.cos(np.radians(selected.rake))*np.sin(np.radians(2*selected.strike)) - (np.sin(np.radians(2*selected.dip))*np.sin(np.radians(selected.rake))*(np.cos(np.radians(selected.strike)))**2))
        Men = (np.sin(np.radians(selected.dip))*np.cos(np.radians(selected.rake))*np.cos(np.radians(2*selected.strike)) + 0.5*np.sin(np.radians(2*selected.dip))*np.sin(np.radians(selected.rake))*np.sin(np.radians(2*selected.strike)))

        ## calculate strain for each fault
        faultEnn = (selected.length_m * (selected.vtotal/1000) * Mnn) / (2*area*np.sin(np.radians(selected.dip)))
        faultEee = (selected.length_m * (selected.vtotal/1000) * Mee) / (2*area*np.sin(np.radians(selected.dip)))
        faultEen = (selected.length_m * (selected.vtotal/1000) * Men) / (2*area*np.sin(np.radians(selected.dip)))

        ## calculate rotated strain for each fault
        faultEee_rotated = faultEee*(np.cos(theta)**2) + faultEnn*(np.sin(theta)**2) + 2*faultEen*np.sin(theta)*np.cos(theta)
        faultEnn_rotated = faultEee*(np.sin(theta)**2) + faultEnn*(np.cos(theta)**2) - 2*faultEen*np.sin(theta)*np.cos(theta)
        faultEen_rotated = (faultEnn-faultEee)*np.sin(theta)*np.cos(theta) + faultEen*((np.cos(theta)**2)-(np.sin(theta)**2))
        selected['faultEen_rotated'] = faultEen_rotated / 1e-9

        # calculated second invariant for each fault
        fault_i2 = np.sqrt((faultEee**2)+(faultEnn**2)+(2*(faultEen**2)))
        selected['fault_i2'] = fault_i2 / 1e-9

        ## sum strain within this bin
        cumulative_Enn = np.sum(faultEnn)
        cumulative_Eee = np.sum(faultEee)
        cumulative_Een = np.sum(faultEen)

        ## sum rotated strain within this bin
        cumulative_Enn_rot = np.sum(faultEnn_rotated)
        cumulative_Eee_rot = np.sum(faultEee_rotated)
        cumulative_Een_rot = np.sum(faultEen_rotated)

        ## assign summed strain to external array for later
        summed_Enn[i] = cumulative_Enn
        summed_Eee[i] = cumulative_Eee
        summed_Een[i] = cumulative_Een

        ## assign rotated summed strain to external array for later
        summed_Enn_rot[i] = cumulative_Enn_rot
        summed_Eee_rot[i] = cumulative_Eee_rot
        summed_Een_rot[i] = cumulative_Een_rot

        ## iterate through selected and assign strain values back to hazfaults
        for index,row in selected.iterrows():
            hazfaults.loc[index,'Eee_rot'] = faultEee_rotated[index]
            hazfaults.loc[index,'Enn_rot'] = faultEnn_rotated[index]
            hazfaults.loc[index,'Een_rot'] = faultEen_rotated[index]


### calculate second invariant
#i2 = np.sqrt((strain.exx**2) + (strain.eyy**2) + (2*(strain.exy**2))) from kreemer et al, 2014
i2 =  np.sqrt((summed_Enn_rot**2) + (summed_Eee_rot**2) + (2*(summed_Een_rot**2))) # binned
hazfaults['i2'] = np.sqrt((hazfaults.Enn_rot**2) + (hazfaults.Eee_rot**2) + (2*(hazfaults.Een_rot**2))) # for each fault

### add summed strain cumulatively from farthest to closest
# first, reverse it
summed_Enn_rot_flip = np.flip(summed_Enn_rot)
summed_Eee_rot_flip = np.flip(summed_Eee_rot)
summed_Een_rot_flip = np.flip(summed_Een_rot)

cumulative_Enn_rot = np.copy(summed_Enn_rot_flip)
cumulative_Eee_rot = np.copy(summed_Eee_rot_flip)
cumulative_Een_rot = np.copy(summed_Een_rot_flip)

# second, add it up
for i in range(len(cumulative_Enn_rot)-1):
        cumulative_Enn_rot[i+1] += cumulative_Enn_rot[i]
        cumulative_Eee_rot[i+1] += cumulative_Eee_rot[i]
        cumulative_Een_rot[i+1] += cumulative_Een_rot[i]

# third, flip it back
cumulative_Enn_rot = np.flip(cumulative_Enn_rot)
cumulative_Eee_rot = np.flip(cumulative_Eee_rot)
cumulative_Een_rot = np.flip(cumulative_Een_rot)

### calculate second invariant cumulative
#i2 = np.sqrt((strain.exx**2) + (strain.eyy**2) + (2*(strain.exy**2))) from Kreemer et al, 2014
i2_cumulative =  np.sqrt((cumulative_Enn_rot**2) + (cumulative_Eee_rot**2) + (2*(cumulative_Een_rot**2)))


############################
#### CALCULATE VELOCITY ####
############################
# see Holt & Haines 1991 appendix and Haines 1982

# convert bin width from m to mm so we get results in mm/yr
y = width * 1000

## calculate parallel and normal velocity for each bin
# u direction is parallel to bin (bin length); v direction is normal to bin (bin width)
# we've assumed Enn (extensional strain parallel to strike) is negligible
u = 2*summed_Een_rot * y
v = summed_Eee_rot * y

## calculate cumulative velocities
cumulative_u = np.copy(np.flip(u))
cumulative_v = np.copy(np.flip(v))

for i in range(len(cumulative_u)-1):
        cumulative_u[i+1] += cumulative_u[i]
        cumulative_v[i+1] += cumulative_v[i]

cumulative_u = np.flip(cumulative_u)
cumulative_v = np.flip(cumulative_v)

## calculate rotation for each bin
w = np.zeros(len(u),dtype=float)

for i in range(len(u)-1):
    w[i] = 0.5 * (u[i+1]-u[i]) / y

w = np.degrees(w) # convert from radians to degrees

print('done')


#####################
#### SAVE OUTPUT ####
#####################
### save strain ###
with open('strain_results_cumulative_%s_%skm.p'%(strike,step),'wb') as f:
    pickle.dump((cumulative_Enn_rot,cumulative_Eee_rot,cumulative_Een_rot), f)

with open('strain_results_binned_%s_%skm.p'%(strike,step),'wb') as f:
    pickle.dump((summed_Enn_rot,summed_Eee_rot,summed_Een_rot), f)

### save velocity ###
with open('velocity_results_%s_%skm.p'%(strike,step),'wb') as f:
    pickle.dump((u,v,cumulative_u,cumulative_v), f)

### save bins, lengths, areas ###
with open('bins_%s_%skm.p' %(strike,step),'wb') as f:
    pickle.dump((bins,lengths,areas),f)
